# Getty AAT
# A module to parse query results (json) from The Getty Art & Archiechture Thesaurus (AAT)

import json
import gzip
import re
import csv
from SPARQLWrapper import SPARQLWrapper, JSON
import nltk
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer
import simplemma
import pandas as pd
import requests
import bows

def main():
	if __name__ == "__main__":
		main()

def sparql(aat_uri:list, lang:str) -> dict:
    '''
    Querying prefLabel, altLabel, scopeNote, rdfs comments of concepts in AAT;
    Sends SPARQL queries to the AAT endpoint via SPARQLwrapper 
    aat_uri: list of AAT concepts IDs (str) ['ID']
    lang: str 'en' or 'nl'
    Returns a dict with query results: {'ID':{'lang':'en',
                                              'prefLabel':'',
                                              'altLabels':[],
                                              'scopeNote':'',
                                              'prefLabel_comment':'',
                                              'altLabel_comment':''}
    '''
    
    sparql = SPARQLWrapper("http://vocab.getty.edu/sparql")
    
    if lang == 'en':
        lang_code = '300388277'
    if lang == 'nl':
        lang_code = '300388256'
        
    result_dict = {}
        
    for uri in aat_uri:
        
        result_dict[uri] = {}
        
        query_string = '''SELECT ?prefLabel (GROUP_CONCAT(?altLabel;SEPARATOR="#") AS ?altLabels)
        ?scopeNote ?prefLabel_comment ?altLabel_comment
        WHERE {aat:''' + uri + ''' xl:prefLabel ?pL .?pL dcterms:language aat:''' + lang_code + ''';
        xl:literalForm ?prefLabel .
        OPTIONAL {?pL rdfs:comment ?prefLabel_comment . }
        OPTIONAL {aat:''' + uri + ''' xl:altLabel ?aL .
        ?aL dcterms:language aat:''' + lang_code + ''';
        xl:literalForm ?altLabel . 
        OPTIONAL { ?aL rdfs:comment ?altLabel_comment . }}
        OPTIONAL {aat:''' + uri + ''' skos:scopeNote / dcterms:language aat:'''+ lang_code + ''';
        skos:scopeNote / rdf:value ?scopeNote . }}
        GROUP BY ?prefLabel ?scopeNote ?prefLabel_comment ?altLabel_comment'''
        
        sparql.setQuery(query_string)
        sparql.setReturnFormat(JSON)
        results = sparql.query().convert()
        
        prefLabel = None
        altLabels = []
        scopeNote = None
        prefLabel_comment = None
        altLabel_comment = None
        
        for result in results['results']['bindings']:
            
            if 'prefLabel' in result:
                prefLabel = result['prefLabel']['value']
            if 'altLabels' in result:
                altLabels = result['altLabels']['value'].split('#')
            if 'scopeNote' in result:
                scopeNote = result['scopeNote']['value']
            if 'prefLabel_comment' in result:
                prefLabel_comment = result['prefLabel_comment']['value']
            if 'altLabel_comment' in result:
                altLabel_comment = result['altLabel_comment']['value']

            result_dict[uri]['lang'] = lang
            result_dict[uri]['prefLabel'] = prefLabel
            result_dict[uri]['altLabels'] = altLabels
            result_dict[uri]['prefLabel_comment'] = prefLabel_comment
            result_dict[uri]['altLabel_comment'] = altLabel_comment
            result_dict[uri]['scopeNote'] = scopeNote
        
    return result_dict


def _get_entity_info(entity_id:str, lang:str, aat_gzip:dict) -> dict:
    '''
    This function is used for 'find_term_in_literals'
    Getting the values of prefLabel (str), altLabel (list), prefLabel_comment (str), altLabel_comment(list), and scopeNote (str)
    by entity ID in AAT
    entity_id: str, entity ID in AAT (for example, '300189559')
    lang: str, 'en' or 'nl'; language of entities info
    aat_gzip: dict, ungzipped json file with search results (ungzips in the main function)
    Returns a dict {'entity': '',
					 'scopeNote': '',
					 'prefLabel_comment': '',
					 'prefLabel': '',
					 'altLabel': [],
					 'altLabel_comment': []}
    '''

    results = {}
    results['entity'] = entity_id
    altLabel_list = []
    altLabel_comment_list = []
    results['scopeNote'] = ''
    results['prefLabel_comment'] = ''
    
    for triple in aat_gzip["results"]["bindings"]:
        if entity_id in triple['Subject']['value']:
            
            # prefLabel
            if 'prefLabel' in triple['Predicate']['value']:
                for triple_t in aat_gzip["results"]["bindings"]:
                    if triple_t['Subject']['value'] == triple['Object']['value']:
                        if 'literalForm' in triple_t['Predicate']['value']:
                            results['prefLabel'] = triple_t['Object']['value']
                        # prefLabel comment
                        if 'comment' in triple_t['Predicate']['value']:
                            results['prefLabel_comment'] = triple_t['Object']['value']
                        
            # altLabel
            if 'altLabel' in triple['Predicate']['value']:
                for triple_t in aat_gzip["results"]["bindings"]:
                    if triple_t['Subject']['value'] == triple['Object']['value']:
                        if 'literalForm' in triple_t['Predicate']['value']:
                            altLabel_list.append(triple_t['Object']['value'])
                        # altLabel comment
                        if 'comment' in triple_t['Predicate']['value']:
                            altLabel_comment_list.append(triple_t['Object']['value'])
                            
            # scopeNote
            if 'scopeNote' in triple['Predicate']['value']:
                for triple_t in aat_gzip["results"]["bindings"]:
                    if triple_t['Subject']['value'] == triple['Object']['value']:
                        results['scopeNote'] = triple_t['Object']['value']
                            
            results['altLabel'] = altLabel_list
            results['altLabel_comment'] = altLabel_comment_list
                        
    return results

def find_term_in_literals(query_term:str, lang:str) -> list:
    '''
    Finding query terms in the literal values of properties:
    prefLabel, altLabel, rdfs comment (for prefLabel and altLabel), and scopeNote
    query_term: str, for example 'term'
    ang: str, 'en' or 'nl'; language of the literals in which the query term is searched
    Returns a list of dicts with an AAT entity URI and the property name in the literal value of which the term was found
    '''

    # reading the gzip json file with aat search results
    # path to raw gzip on GitHub
    gzip_path = f"https://github.com/cultural-ai/LODlit/raw/main/AAT/gzip_aat_subgraph_{lang}.json"
    
    # decompressing
    aat_gzip = json.loads(gzip.decompress(requests.get(gzip_path).content))
    
    list_of_results = []
    
    for triple in aat_gzip["results"]["bindings"]:
        
        if triple['Object']['type'] == 'literal' \
        and len(re.findall(f'\\b{query_term}\\b',triple['Object']['value'],re.IGNORECASE)) > 0:
            
            results_per_hit = {}
            results_per_hit['query_term'] = query_term
            results_per_hit['aat_uri'] = ''
            
            # if a term found in scopeNote
                    
            if 'rdf-syntax' in triple['Predicate']['value']:
                results_per_hit['found_in'] = 'scopeNote'
                for triple_t in aat_gzip["results"]["bindings"]:
                    if triple_t['Object']['value'] == triple['Subject']['value']:
                        # getting entity URI
                        entity = triple_t['Subject']['value'].split('/')[-1]
                        
            # if a term found in pref or alt labels
            
            if 'literalForm' in triple['Predicate']['value']:
                for triple_t in aat_gzip["results"]["bindings"]:
                    if triple_t['Object']['value'] == triple['Subject']['value']:
                        # getting entity URI
                        entity = triple_t['Subject']['value'].split('/')[-1]
                        # altLabel or prefLabel 
                        if 'altLabel' in triple_t['Predicate']['value']:
                            results_per_hit['found_in'] = 'altLabel'
                        if 'prefLabel' in triple_t['Predicate']['value']:
                            results_per_hit['found_in'] = 'prefLabel'
                        
            # if a term found in rdfs comment
            
            if 'comment' in triple['Predicate']['value']:
                for triple_t in aat_gzip["results"]["bindings"]:
                    if triple_t['Object']['value'] == triple['Subject']['value']:
                        # getting entity URI
                        entity = triple_t['Subject']['value'].split('/')[-1]
                        # comment to altLabel or prefLabel
                        if 'altLabel' in triple_t['Predicate']['value']:
                            results_per_hit['found_in'] = 'altLabel_comment'
                        if 'prefLabel' in triple_t['Predicate']['value']:
                            results_per_hit['found_in'] = 'prefLabel_comment'
            
            results_per_hit['aat_uri'] = entity
            
            entity_info = _get_entity_info(entity, lang, aat_gzip)
            
            results_per_hit['prefLabel'] = entity_info['prefLabel']
            results_per_hit['prefLabel_comment'] = entity_info['prefLabel_comment']
            results_per_hit['altLabel'] = entity_info['altLabel']
            results_per_hit['altLabel_comment'] = entity_info['altLabel_comment']
            results_per_hit['scopeNote'] = entity_info['scopeNote']
            
            list_of_results.append(results_per_hit)
            
    return list_of_results

def get_bows(lang:str) -> dict:
    '''
    Getting bag of words (BoW) from the AAT search results for every search term
    lang: str, 'en' or 'nl'
    Returns a dict with BoWs per hit per term: {term:[{aat_URI:['token1','token2','token3']}]}
    '''

    path_to_results = f"https://github.com/cultural-ai/LODlit/raw/main/AAT/aat_query_results_{lang}.json"
    search_results = requests.get(path_to_results).json()

    wnl = WordNetLemmatizer()
    all_bows = {}

    for query_term, results in search_results.items():

        list_by_term = []

        for hit in results:
            q_bag = {}
            literals = []
            
            literals.append(hit["prefLabel"])
            literals.append(hit["prefLabel_comment"])
            literals.append(hit["scopeNote"])
            literals.extend(hit["altLabel"])
            literals.extend(hit["altLabel_comment"])

            bow = bows.make_bows(literals,lang,merge_bows=True)

            q_bag[hit["aat_uri"]] = bow

            list_by_term.append(q_bag)

        all_bows[query_term] = list_by_term

    return all_bows


def get_lit_related_matches_bow(lang:str) -> dict:
    '''
    Generates dict with BoWs by query terms and their related matches in AAT
    reads files with AAT BoWs: aat_bows_en.json (EN), aat_bows_nl.json (NL)
    the info about which AAT concept is a related match to the query term is taken from https://github.com/cultural-ai/wordsmatter/blob/main/related_matches/rm.csv
    lang: str, 'en' or 'nl'
    Returns a dict with BoWs by term: {term:{concept_uri:['token1','token2','token3']}}
    '''
    
    results = {}
    
    # loading related matches from GitHub
    path_rm = "https://github.com/cultural-ai/wordsmatter/raw/main/related_matches/rm.json"
    rm = requests.get(path_rm).json()
    
    # load aat bows
    path_to_bows = f"https://github.com/cultural-ai/LODlit/raw/main/AAT/aat_bows_{lang}.json"
    aat_bows = requests.get(path_to_bows).json()

    # getting a list of all AAT URIs of related matches
    # terms with no related matches won't be included in the file
    related_matches_aat = list(set([values["related_matches"]["aat"][0] for values in rm.values() \
                            if values["lang"] == lang and values["related_matches"]["aat"][0] != 'None']))    
    
    # getting BoWs for AAT concept URIs 
    related_matches_aat_uri_bows = {}
    for uri_rm in related_matches_aat:
        for hits in aat_bows.values():
            for hit in hits:
                for uri, bow in hit.items():
                    if uri == uri_rm:
                        related_matches_aat_uri_bows[uri_rm] = bow
                        
    # shaping resulting dict: terms with related matches and BoWs
    for values in rm.values():
        if values["lang"] == lang:
            rm_aat = values["related_matches"]["aat"][0]
            if rm_aat != "None":
                for term in values["query_terms"]:
                    if rm_aat in related_matches_aat_uri_bows.keys():
                        results[term] = {"aat_uri":rm_aat,"bow":related_matches_aat_uri_bows[rm_aat]}
            
    return results


def get_cs(lang:str):
    '''
    Calculating three cosine similarity scores between AAT search results and background info;
    the similarity scores are based on (1) only related matches, (2) only WM text, and (3) extended bows with related matches and WM text
    lang:str, language of bows, "en" or "nl"
    Returns a pandas data frame with columns:
    query_term, aat_URI, bow, cs_rm, cs_wm, cs_rm_wm
    '''

    nlp = bows.load_spacy_nlp(lang)

    # load bckground info
    path_bg = "https://github.com/cultural-ai/LODlit/raw/main/bg/background_info_bows.json"
    bg_info = requests.get(path_bg).json()

    # load aat bows

    path_aat_bows = f"https://github.com/cultural-ai/LODlit/raw/main/AAT/aat_bows_{lang}.json"
    aat_bows = requests.get(path_aat_bows).json()

    aat_df = pd.DataFrame(columns=['term','hit_id','bow','cs_rm','cs_wm','cs_rm_wm'])

    for term, hits in aat_bows.items():

        bg_rm = bows._collect_bg(term,lang,bg_info,bg_bow="rm")
        bg_wm = bows._collect_bg(term,lang,bg_info,bg_bow="wm")
        bg_rm_wm = bows._collect_bg(term,lang,bg_info,bg_bow="joint")
          
        # if there are search results
        if len(hits) > 0:
            for hit in hits:
                for i, bow in hit.items():
                    # making a set
                    aat_bow = list(set(bow))
                    
                    if len(aat_bow) > 0:
                        # calculate cs
                        cs_rm = bows.calculate_cs(bg_rm,aat_bow,nlp)
                        cs_wm = bows.calculate_cs(bg_wm,aat_bow,nlp)
                        cs_rm_wm = bows.calculate_cs(bg_rm_wm,aat_bow,nlp)
                        
                        aat_df.loc[len(aat_df)] = [term,i,aat_bow,cs_rm,cs_wm,cs_rm_wm]
                    
                    # if there are no tokens, all cs = None
                    else:
                        aat_df.loc[len(aat_df)] = [term,i,aat_bow,None,None,None]

        # if there are no search results in AAT, cs = None
        else:
            aat_df.loc[len(aat_df)] = [term,None,None,None,None,None]

    return aat_df


# Module to parse Princeton WordNet 3.1

# Important! After installing NLTK, download wordnet31 ('https://github.com/nltk/nltk_data/blob/gh-pages/packages/corpora/wordnet31.zip');
# put the content of 'wordnet31' to 'wordnet' in 'nltk_data/corpora' (it is not possible to import wordnet31 from nltk.corpus; See explanations on the WordNet website (retrieved on 10.02.2023): https://wordnet.princeton.edu/download/current-version; "WordNet 3.1 DATABASE FILES ONLY. You can download the WordNet 3.1 database files. Note that this is not a full package as those above, nor does it contain any code for running WordNet. However, you can replace the files in the database directory of your 3.0 local installation with these files and the WordNet interface will run, returning entries from the 3.1 database. This is simply a compressed tar file of the WordNet 3.1 database files.";
# Use check_version() to ensure that WordNet 3.1 is imported

import json
import re
import requests
import nltk
from nltk.corpus import wordnet as wn
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer
import warnings
import pandas as pd
import bows

def main():

    print("Install NLTK and download wordnet. Download wordnet31 from 'https://github.com/nltk/nltk_data/blob/gh-pages/packages/corpora/wordnet31.zip' and put the content of 'wordnet31' to 'wordnet' in 'nltk_data/corpora'")

    if __name__ == "__main__":
        main()

def check_version() -> str:
    return f"You have WordNet version {wn.get_version()}"

def find_term(query_term:str) -> list:
    '''
    Searching for a query term in synset lemmas, synset difinitions, and synset examples
    query_term: str, a term to search in synsets, phrases are also allowed
    Returns a list of dicts with the synset info if the query term was found:
    [{'query_term':query_term,
        'synset_id':'',
        'lemmata':[],
        'definition':'',
        'examples':[],
        'found_in':''}]
    '''

    if wn.get_version() != '3.1':
        warnings.warn("This module is created to parse WordNet 3.1. Your version is different. Download version 3.1 from 'https://github.com/nltk/nltk_data/blob/gh-pages/packages/corpora/wordnet31.zip' and put the content of 'wordnet31' to 'wordnet' in 'nltk_data/corpora'. Import wordnet again and check your version with .get_version()")

    search_results = []
    # searching in lemmata
    # getting synset_id, lemmata (synonyms), definition, examples
    for synset in wn.synsets(query_term):
        for le in synset.lemmas():
            # exact match between query term and lemma name
            if query_term == le.name().lower(): # lemmas can be capitalized
                result_dict = {}
                result_dict['query_term'] = query_term
                result_dict['synset_id'] = synset.name()
                result_dict['lemmata'] = [l.name() for l in synset.lemmas()]
                result_dict['definition'] = synset.definition()
                result_dict['examples'] = synset.examples()
                result_dict['found_in'] = 'lemmata'
                search_results.append(result_dict)

    # searching in all synsets
    for synset in list(wn.all_synsets()):

        # searching in definitions
        if len(re.findall(f'\\b{query_term}\\b',synset.definition(),re.IGNORECASE)) > 0:
            result_dict = {}
            result_dict['query_term'] = query_term
            result_dict['synset_id'] = synset.name()
            result_dict['lemmata'] = [l.name() for l in synset.lemmas()]
            result_dict['definition'] = synset.definition()
            result_dict['examples'] = synset.examples()
            result_dict['found_in'] = 'definition'
            search_results.append(result_dict)

        # searching in examples
        for example in synset.examples():
            if len(re.findall(f'\\b{query_term}\\b',example,re.IGNORECASE)) > 0:
                result_dict = {}
                result_dict['query_term'] = query_term
                result_dict['synset_id'] = synset.name()
                result_dict['lemmata'] = [l.name() for l in synset.lemmas()]
                result_dict['definition'] = synset.definition()
                result_dict['examples'] = synset.examples()
                result_dict['found_in'] = 'examples'
                search_results.append(result_dict)

    return search_results

def get_synsets_info(synsets:list) -> dict:
    '''
    Getting lemmata, definition, and examples of synsets
    synsets: a list of synsets IDs (str), for example ['cookie.n.01','cookie.n.02','cookie.n.03']
    Return a dict: {'synset_id': {'lemmata': [],
                                   'definition': '',
                                   'examples': []}
    '''
    
    synsets_info = {}
    
    for s in synsets:
        synset = wn.synset(s)
        lemmata = [l.name() for l in synset.lemmas()]
        definition = synset.definition()
        examples = synset.examples()
        
        # writing results 
        synsets_info[s] = {'lemmata': lemmata,
                         'definition': definition,
                         'examples': examples}
        
    return synsets_info

def get_bows(path_to_results:str) -> dict:
    '''
    Getting bag of words (BoW) from the Princeton WordNet 3.1 search results for every search term
    path_to_results: str, a path to the search results (in json format)
    Returns a dict with BoWs per hit per term: {term:[{synset_id:['token1','token2','token3']}]}
    '''
    wnl = WordNetLemmatizer()

    all_bows = {}
   
    with open(path_to_results,'r') as jf:
        search_results = json.load(jf)

    for query_term, results in search_results.items():

        list_by_term = []

        for hit in results:

            q_bag = {}
            literals = []

            literals.extend(hit["lemmata"])
    
            if hit["definition"] != "":
                literals.append(hit["definition"])
        
            literals.extend(hit["examples"])

            bow = bows.make_bows(literals,"en",merge_bows=True)

            q_bag[hit["synset_id"]] = bow

            list_by_term.append(q_bag)

        all_bows[query_term] = list_by_term

    return all_bows


def get_lit_related_matches_bow() -> dict:
    '''
    Generates dict with BoWs by query terms and their related matches in PWN
    reads the file with PWN BoWs: pwn31_bows.json
    the info about which PWN synset is a related match to the query term is taken from https://github.com/cultural-ai/wordsmatter/blob/main/related_matches/rm.csv
    Returns a dict with BoWs by term: {term:['token1','token2','token3']}
    This dict doesn't store info about synsets IDs bc one term can have one or more related match synsets;
    in these cases, BoWs of each synsets are merged
    '''
    
    results = {}
    
    # loading related matches
    path_rm = "https://github.com/cultural-ai/wordsmatter/raw/main/related_matches/rm.json"
    rm = requests.get(path_rm).json()

    # load pwn bows
    path_to_bows = f"https://github.com/cultural-ai/LODlit/raw/main/PWN/pwn31_bows.json"
    pwn_bows = requests.get(path_to_bows).json()

    # getting a list of all PWN synsets of related matches
    pwn_synsets = []
    for values in rm.values():
        if values["lang"] == "en" and values["related_matches"]["pwn"][0] != 'None':
            pwn_synsets.extend(values["related_matches"]["pwn"])

    related_matches_pwn = list(set(pwn_synsets))

    # getting BoWs for each synset 
    related_matches_pwn_synsets_bows = {}
    for s_rm in related_matches_pwn:
        for hits in pwn_bows.values():
            for hit in hits:
                for s, bow in hit.items():
                    if s == s_rm:
                        related_matches_pwn_synsets_bows[s_rm] = bow

    # shaping resulting dict: BoWs per term
    for values in rm.values():
        if values["lang"] == "en":
            rm_synsets = values["related_matches"]["pwn"]
            if rm_synsets[0] != "None":
                for term in values["query_terms"]:
                    synsets_bow = []
                    for s in rm_synsets:
                        if s in related_matches_pwn_synsets_bows.keys():
                            synsets_bow.extend(related_matches_pwn_synsets_bows[s])
                            results[term] = synsets_bow

    return results

def get_cs():
    '''
    Calculating three cosine similarity scores between PWN search results and background info;
    the similarity scores are based on (1) only related matches, (2) only WM text, and (3) extended bows with related matches and WM text
    Returns a pandas data frame with columns:
    query_term, hit_id, bow, cs_rm, cs_wm, cs_rm_wm
    '''

    nlp = bows.load_spacy_nlp("en")

    # load bckground info
    path_bg = "https://github.com/cultural-ai/LODlit/raw/main/bg/background_info_bows.json"
    bg_info = requests.get(path_bg).json()

    # load all pwn bows
    path_pwn_bows = f"https://github.com/cultural-ai/LODlit/raw/main/PWN/pwn31_bows.json"
    pwn_bows = requests.get(path_pwn_bows).json()

    pwn_df = pd.DataFrame(columns=['term','hit_id','bow','cs_rm','cs_wm','cs_rm_wm'])

    for term, hits in pwn_bows.items():

        bg_rm = bows._collect_bg(term,"en",bg_info,bg_bow="rm")
        bg_wm = bows._collect_bg(term,"en",bg_info,bg_bow="wm")
        bg_rm_wm = bows._collect_bg(term,"en",bg_info,bg_bow="joint")

        # if there are search results
        if len(hits) > 0:
            for hit in hits:
                for i, bow in hit.items():
                    # making a set
                    pwn_bow = list(set(bow))

                    if len(pwn_bow) > 0:
                        # calculate cs
                        cs_rm = bows.calculate_cs(bg_rm,pwn_bow,nlp)
                        cs_wm = bows.calculate_cs(bg_wm,pwn_bow,nlp)
                        cs_rm_wm = bows.calculate_cs(bg_rm_wm,pwn_bow,nlp)
                        
                        pwn_df.loc[len(pwn_df)] = [term,i,pwn_bow,cs_rm,cs_wm,cs_rm_wm]
                    
                    # if there are no tokens, all cs = None
                    else:
                        pwn_df.loc[len(pwn_df)] = [term,i,pwn_bow,None,None,None]

        # if there are no search results in PWN, cs = None
        else:
            pwn_df.loc[len(pwn_df)] = [term,None,None,None,None,None]

    return pwn_df


# Functions to make bag-of-words from strings, collect background information, and  calculare cosine similarity
# To use 'calculate_cs', download 'en_core_web_lg' and 'nl_core_news_lg' from https://spacy.io/models:
# python -m spacy download en_core_web_lg AND python -m spacy download nl_core_news_lg

import json
import re
import requests
import math
import pandas as pd
import nltk
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer
import simplemma
import spacy # numpy version 1.21

def main():
    if __name__ == "__main__":
        main()

def make_bows(text:list, lang:str, merge_bows=False) -> list:
    '''
    Makes a BoW from a list of str:
    removes non-word charachters (incl. punctuation, numbers),
    removes stop-words (nltk),
    lowercases, tokenises (split by space), lemmatises (NLTK lemmatiser for EN; simplemma for NL),
    removes tokens with fewer than 3 characters;
    text: a list of str to make BoWs from
    lang: str, language of strings, 'en' or 'nl'
    merge_bows: bool, if there are multiple texts, merge them in one BoW (True) or not (False), default False 
    Returns a BoW (list of lists)
    '''
    
    wnl = WordNetLemmatizer()

    bows = []
    
    for t in text:
        no_w_text = re.sub('(\W|\d)',' ',t)
        text_bow = no_w_text.split(' ')
        
        # checking lang
        if lang == 'en':
            text_bow_clean = [wnl.lemmatize(token.lower()) for token in text_bow if token.lower() not in stopwords.words('english') and token != '' and len(token) > 2]
        if lang == 'nl':
            # Dutch lemmatizer can output uppercase lemmas
            text_bow_clean = [simplemma.lemmatize(token.lower(),lang='nl').lower() for token in text_bow if token.lower() not in stopwords.words('dutch') and token != '' and len(token) > 2]
            
        if merge_bows == True:
            bows.extend(text_bow_clean)
        else:
            bows.append(text_bow_clean)
            
    return bows


def tf(doc:list, token:str) -> float:
    '''
    Calculates term frequency (TF)
    doc: list, all documents
    token: str, a token to get the TF score of
    Returns float
    '''
    n_found = len([t for t in doc if t == token])
    tf_score = n_found / len(doc)
    
    return tf_score


def idf(token:str, doc_freq:dict, n_docs:int) -> float:
    '''
    Calculates inverse document frequency (IDF):
        adds 1 to DF to avoid zero division
    token: str, a token to get the IDF score of
    doc_freq: dict, document frequency, in how many documents tokens appear
    n_docs: int, a number of documents
    Returns float
    '''
    idf_score = math.log(n_docs / (doc_freq[token] + 1))
        
    return idf_score


def get_top_tokens_tfidf(bow:list,doc_freq:dict,n_docs:int) -> list:
    '''
    Getting top tokens based on their TF-IDF weighting in one BoW
    Depends on the tf and idf functions
    bow: list of str, tokens in one BoW
    doc_freq: dict, document frequency, in how many documents tokens appear ({'token':int})
    n_docs: int, a number of documents
    Returns list: top 10 tokens in a bow by their TF-IDF scores
    '''
    top_tokens = []
    tf_idf_scores = {}
    
    for token in bow:
        tf_idf = tf(bow,token) * idf(token,doc_freq,n_docs)
        tf_idf_scores[token] = tf_idf
        
    tokens_scores = sorted(tf_idf_scores.items(), key=lambda x:x[1], reverse=True)
    
    if len(tokens_scores) < 10:
        top_tokens = [t[0] for t in tokens_scores]
    else:
        top_tokens = [t[0] for t in tokens_scores[0:10]]
    
    return top_tokens

def load_spacy_nlp(lang:str):
    '''
    Loads an NLP pipeline from spacy for terms vectorisation;
    lang: str, language of the strings to process, 'en' or 'nl';
    before loading, download 'en_core_web_lg' and 'nl_core_news_lg' from https://spacy.io/models
    
    '''
    # check lang
    if lang == "en":
        nlp = spacy.load("en_core_web_lg", disable=["tagger", "attribute_ruler", "lemmatizer"])

    # load nl model    
    if lang == "nl":
        nlp = spacy.load("nl_core_news_lg", disable=["tagger", "attribute_ruler", "lemmatizer"])

    return nlp

def calculate_cs(bow_1:list, bow_2:list, nlp) -> float:
    '''
    Calculates cosine similarity between two bags of words;
    based on the spacy vectors
    bow_1 and bow_2: list, two bags of words;
    nlp: spacy nlp class loaded through spacy.load; use the function 'load_spacy_nlp'
    '''

    # converting list to str, spaces as a separator

    bow_1_str = ""
    for t in bow_1:
        bow_1_str += f"{t} "
        
    bow_2_str = ""
    for t in bow_2:
        bow_2_str += f"{t} "
        
    bow_1 = nlp(bow_1_str)
    bow_2 = nlp(bow_2_str)
        
    sim = bow_1.similarity(bow_2)
    
    return sim


def _collect_bg(term:str, lang:str, bg_info:dict, bg_bow:str) -> list:
    '''
    Collects bacgkround info for a term
    term: str, a term to get backgrounf info for
    lang: str, lang of the term, "en" or "nl"
    bg_info: dcit with background info (loads in "get_cs" functions)
    bg_bow: str, indicates which background info to collect;
        options for bg_bow: (1) "rm" -- related matches from all resources, not including Words Matter texts, (2) "wm" -- only Words Matter text, (3) "joint" -- related matches + Words Matter
    Returns a list of unique tokens from the background info of a term
    '''

    if term in bg_info[lang].keys():
        bg = []

        if bg_bow == "rm":

            if bg_info[lang][term].get("aat"):
                bg.extend(bg_info[lang][term]["aat"])
            
            if bg_info[lang][term].get("wikidata"):
                bg.extend(bg_info[lang][term]["wikidata"])

            if lang == "en":
                if bg_info[lang][term].get("pwn"):
                    bg.extend(bg_info[lang][term]["pwn"])
            if lang == "nl":
                if bg_info[lang][term].get("odwn"):
                    bg.extend(bg_info[lang][term]["odwn"])

        if bg_bow == "wm":

            if bg_info[lang][term].get("wm"):
                bg.extend(bg_info[lang][term]["wm"])

        if bg_bow == "joint":

            if bg_info[lang][term].get("aat"):
                bg.extend(bg_info[lang][term]["aat"])
            
            if bg_info[lang][term].get("wikidata"):
                bg.extend(bg_info[lang][term]["wikidata"])

            if lang == "en":
                if bg_info[lang][term].get("pwn"):
                    bg.extend(bg_info[lang][term]["pwn"])
            if lang == "nl":
                if bg_info[lang][term].get("odwn"):
                    bg.extend(bg_info[lang][term]["odwn"])

            if bg_info[lang][term].get("wm"):
                bg.extend(bg_info[lang][term]["wm"])
            
        bg_set = list(set(bg))

    else:
        bg_set = []

    return bg_set

def _vectorize_bows(bow_1:list, bow_2:list) -> int:
    '''
    Vectorizing two BoWs
    based on common token overlap in two sets of tokens
    Calculating cosine similarity between them
    bow_1 and bow_2: list, bag of words
    Returns int, rounded cosine similarity value
    '''
    # creating sets
    set_1 = {t for t in bow_1}
    set_2 = {t for t in bow_2}
    union_set = set_1.union(set_2)
    
    # vectorizing
    v1 = []
    v2 = []
    for t in union_set:
        if t in set_1:
            v1.append(1)
        else:
            v1.append(0)
        if t in set_2:
            v2.append(1)
        else:
            v2.append(0)
    
    # calculating cs 
    c = 0
    for i in range(len(union_set)):
        c += v1[i] * v2[i]

    cos_sim = c / float((sum(v1) * sum(v2)) ** 0.5)
    
    return cos_sim


def get_top_10(cs_table_path:str, metric:str, lang:str, groupby='lemma'):
    '''
    Getting top-10 results (max 10) (entities) for every query term based on a metric
    cs_table_path: str, path to the csv table with cosine similarity scores, see the files '[dataset]_[lang]_cs.csv'; for example, 'aat_en_cs.csv'; wikidata en cs is zipped;
    metric: str, which metric to take to get top-10
        options for metric: (1) "cs_rm" -- only related matches, (2) "cs_wm" -- only WM text, and (3) "cs_rm_wm" -- extended bows with related matches and WM text
    lang: str, 'en' or 'nl'; language of the dataset in teh csv table 
    groupby: str, 'lemma' or 'term'; how to group entities to generate Top-10; 'lemma' is default  
    Returns a pandas data frame (a subset of cs_table) with ranked results per term or per lemma
    '''

    top_10 = pd.DataFrame()
    cs_table = pd.read_csv(cs_table_path)

    if groupby == 'term':
        # dropping duplicates (taking only unique entities for top-10)
        cs_table.drop_duplicates(subset=["term","hit_id"], inplace=True)

        for group in cs_table.groupby("term"):
            top_10 = top_10.append(group[1].sort_values(by=metric, ascending=False)[0:10])

    if groupby == 'lemma':

        # loading query terms
        path_query_terms = "https://github.com/cultural-ai/LODlit/raw/main/query_terms.json"
        query_terms = requests.get(path_query_terms).json()

        # insert the lemmas column
        lemmas = []
        for row in cs_table.iterrows():
            for lemma, wordforms in query_terms[lang].items():
                if row[1]['term'] in wordforms:
                    lemmas.append(lemma)

        cs_table.insert(0,"lemma",lemmas)

        # drop duplicates
        cs_table.drop_duplicates(subset=["lemma","hit_id"], inplace=True)

        for group in cs_table.groupby("lemma"):
            top_10 = top_10.append(group[1].sort_values(by=metric, ascending=False)[0:10])

    return top_10




# Wikidata module: sending API requests and claning the search results

import nltk
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer
import simplemma
import json
import requests
import time
import re
import pandas as pd
import gzip
import warnings
import bows

def main():
    if __name__ == "__main__":
        main()

def get_search_hits(search_term:str,lang:str,user_agent:str,filter_by_keywords=[],filter_by_statements=[],stemming=False) -> dict:
    """
    Getting search hits of a term in Wikidata
    search_term: str a term or a phrase to search in Wikidata
    lang: str with language code; for example, 'en' or 'nl';
    # (see language codes in Wikidata: https://www.wikidata.org/w/api.php?action=help&modules=wbgetentities)
    user_agent: str with user-agent's info; required by Wikidata (see: https://meta.wikimedia.org/wiki/User-Agent_policy)
    filter_by_keywords: list optional;  the keywords to exclude from the search results; default is []
    filter_by_statements: list optional; list of tuples with statements to exclude from the search results, specifying property and its value like [('P31','Q5'),('P31', 'Q16521')]
    # (meaning that entities with the property P31 ("instance of") with the values "Q5" ("human") and "Q16521" ("taxon") will be excluded);
    # multiple statements are possible, separate them with a whitespace; default is ''
    stemming: bool is to perform stemming of the search term (works only for English); default is False
    Returns a dict
    """

    search_hits = {}

    # 'query' with 'search' generator: constant params
    url = "https://www.wikidata.org/w/api.php"
    params = {"action":"query",
              "prop":"entityterms",
              "wbetlanguage":lang,
              "generator":"search",
              "gsrsearch":f'"{search_term}"', # term goes here (quotes for stemming off)
              "gsrlimit":"1", # 1 result is enough to get the hits info
              "gsrinfo":"totalhits",
              "format":"json"}
    headers = {"user-agent":user_agent}

    # stemming
    if stemming == True:
        params["gsrsearch"] = search_term

    # filtering by keywords
    if filter_by_keywords != []:
        keyword_str = ''
        for kw in filter_by_keywords:
            keyword_str = keyword_str + f" -{kw}"
        params["gsrsearch"] = params["gsrsearch"] + keyword_str

    # filtering by statements    
    if filter_by_statements != []:
        statements_str = ''
        for t in filter_by_statements:
            statements_str = statements_str + f" -haswbstatement:{t[0]}={t[1]}"
        params["gsrsearch"] = params["gsrsearch"] + statements_str

    r = requests.get(url,params=params,headers=headers)
    hits = r.json()['query']['searchinfo']['totalhits']
    search_hits[search_term] = hits

    return search_hits


def get_lit_by_term(search_term:str,lang:str,user_agent:str,filter_by_keywords=[],filter_by_statements=[],stemming=False) -> dict:
    """
    Getting Qids, their labels, aliases, descriptions by a query term; to support stemming, set stemming=True
    search_term: str a term or a phrase to search in Wikidata
    lang: str with language code; for example, 'en' or 'nl';
    # (see language codes in Wikidata: https://www.wikidata.org/w/api.php?action=help&modules=wbgetentities)
    user_agent: str with user-agent's info; required by Wikidata (see: https://meta.wikimedia.org/wiki/User-Agent_policy)
    filter_by_keywords: list optional;  the keywords to exclude from the search results; default is []
    filter_by_statements: list optional; list of tuples with statements to exclude from the search results, specifying property and its value like [('P31','Q5'),('P31', 'Q16521')]
    # (meaning that entities with the property P31 ("instance of") with the values "Q5" ("human") and "Q16521" ("taxon") will be excluded);
    stemming: bool is to perform stemming of the search term (works only for English); default is False
    Returns a dict
    """

    # 'query' with 'search' generator: constant params
    url = "https://www.wikidata.org/w/api.php"
    params = {"action":"query",
              "prop":"entityterms",
              "wbetlanguage":lang,
              "generator":"search",
              "gsrsearch":f'"{search_term}"', # quotes for stemming off
              "gsrlimit":"max", # getting all results
              "gsroffset":"0", # offset
              "gsrinfo":"totalhits",
              "gsrsort":"incoming_links_desc", # sorting results by incoming links
              "format":"json",} 
    headers = {"user-agent":user_agent}

    results = {} # dict to store the results

    # stemming
    if stemming == True:
        params["gsrsearch"] = search_term # stemming on
    
    # filtering by keywords
    if filter_by_keywords != []:
        keyword_str = ''
        for kw in filter_by_keywords:
            keyword_str = keyword_str + f" -{kw}"
        params["gsrsearch"] = params["gsrsearch"] + keyword_str

    # filtering by statements    
    if filter_by_statements != []:
        statements_str = ''
        for t in filter_by_statements:
            statements_str = statements_str + f" -haswbstatement:{t[0]}={t[1]}"
        params["gsrsearch"] = params["gsrsearch"] + statements_str    
    
    # counter for offset
    gsroffset = 0

    # sending requests
    w = requests.get(url,params=params,headers=headers)
    wikidata_json = w.json()
    time.sleep(2) # to prevent 502
        
    # checking the number of hits
    hits = wikidata_json['query']['searchinfo']['totalhits']
        
    # if there are no results
    if hits == 0:
        results[search_term] = wikidata_json['query']
        loops = 0
            
    # saving results from the first loop
    else:
        results[search_term] = wikidata_json['query']['pages']

    # - CONDITIONS FOR LOOPS- #

    # if there are less than 500 hits for a term, no loops are needed
    if hits < 500:
        loops = 0

    # 10K is max; and if hits > 500, offset is needed
    if 10000 > hits > 500 and hits % 500 > 0:
        loops = hits // 500
        
    # minus one loop if there's no remainder
    if 10000 > hits > 500 and hits % 500 == 0:
        loops = hits // 500 - 1
        
    # as the first loop is already done, max = 19
    if hits > 10000:
        loops = 19

    # - REQUEST LOOPS - #   

    for i in range(0,loops):
        gsroffset = gsroffset + 500
        # setting the offset and sending a new request
        params["gsroffset"] = f"{gsroffset}"
        w_i = requests.get(url,params=params,headers=headers)
        wikidata_json_i = w_i.json()
        
        # saving the results
        results[search_term].update(wikidata_json_i['query']['pages'])
        time.sleep(2)

    return results


def get_lit_by_q(qids:list,lang:str,user_agent:str) -> list:
    
    """
    Getting labels, aliases, descriptions of Wikidata entities by their Qids
    qids: list of entity IDs (str) (requests 50 entities at a time slicing the list)
    lang: str with language code; for example, 'en' or 'nl';
    (see language codes in Wikidata: https://www.wikidata.org/w/api.php?action=help&modules=wbgetentities)
    user_agent: str with user-agent's info; required by Wikidata (see: https://meta.wikimedia.org/wiki/User-Agent_policy)
    Returns a list of dicts: [{'QID': {'type': '',
                     'id': 'QID',
                     'labels': {'lang': {'language': 'lang', 'value': ''}},
                     'descriptions': {'lang': {'language': 'lang','value': ''}},
                     'aliases': {'lang': [{'language': 'lang', 'value': ''}]
    """
    
    wd_results = []

    # 'wbgetentities' constant params
    url = "https://www.wikidata.org/w/api.php"
    params = {"action":"wbgetentities",
              "ids":"", # string of entities (max=50)
              "props":"labels|aliases|descriptions",
              "languages":lang,
              "format":"json"}
    headers = {"user-agent":user_agent}

    # - N LOOPS - #

    # if there's a remainder
    if len(qids) % 50 > 0:
        loops = len(qids) // 50 + 1 # add another loop for requests
    else:
        loops = len(qids) // 50

    # - REQUEST LOOPS - #   

    # counters to slice qids

    start_quid_str = 0
    end_quid_str = 0

    for i in range(0,loops):
        ids_string = "" # putting Qs in one string
        end_quid_str = end_quid_str + 50 # max 50 entities per request

        for q in qids[start_quid_str:end_quid_str]:
            ids_string = ids_string + f"{q}|"

        start_quid_str = start_quid_str + 50

        # updating params

        params["ids"] = ids_string.rstrip("|")

        # sending a request
        d = requests.get(url,params=params,headers=headers)
        literals = d.json()

        for qid, value in literals['entities'].items():
            wd_results.append({qid:value})

    return wd_results


def get_labels_by_q(qids:list,lang:list,user_agent:str) -> dict:

    """
    Getting labels of entities by QIDs
    qids: list of entities' QIDs (str) to work on;
    lang: list of str with language code; for example, ['en','nl'];
    (see language codes in Wikidata: https://www.wikidata.org/w/api.php?action=help&modules=wbgetentities)
    user_agent: str with user-agent's info; required by Wikidata (see: https://meta.wikimedia.org/wiki/User-Agent_policy)
    Returns a dict: {"labels":dict with labels by lang,"failed":list of QIDs that were failed}
    """

    # 'wbgetentities' constant params
    url = "https://www.wikidata.org/w/api.php"
    params = {"action":"wbgetentities",
              "ids":"", # string of entities (max=50)
              "props":"labels", # only labels
              "format":"json"}
    headers = {"user-agent":user_agent}

    request = {}
    failed_entities = []

    # languages
    if len(lang) > 1:
        languages = ''
        for l in lang:
            languages = languages + f"{l}|"
        # updating params
        params["languages"] = languages.rstrip("|")
    else:
        params["languages"] = lang[0]

    # - N LOOPS - #

    # if there's a remainder
    if len(qids) % 50 > 0:
        loops = len(qids) // 50 + 1 # add another loop for requests
    else:
        loops = len(qids) // 50

    # - REQUEST LOOPS - #   

    # counters to slice qids

    start_quid_str = 0
    end_quid_str = 0

    for i in range(0,loops):
        ids_string = "" # putting Qs in one string
        end_quid_str = end_quid_str + 50 # max 50 entities per request

        for q in qids[start_quid_str:end_quid_str]:
            ids_string = ids_string + f"{q}|"

        start_quid_str = start_quid_str + 50

        # updating params

        params["ids"] = ids_string.rstrip("|")

        # sending a request
        d = requests.get(url,params=params,headers=headers)
        labels = d.json()

        if 'entities' in labels:
            for entity,lbl in labels['entities'].items():
                per_lang = {}
                for l in lang:
                    if 'labels' in lbl:
                        if l in lbl['labels']:
                            per_lang[l] = lbl['labels'][l]['value']

                request[entity] = per_lang

            # to prevent server errors    
            time.sleep(2)

        else:
            failed_entities.extend(qids[start_quid_str:end_quid_str])

    results = {"labels":request,"failed":failed_entities}

    return results

def _claims_parse_properties(query_results:dict, QID:str, propepties:list) -> dict:
    """
    Parses the values of the properties of entities from Wikidata response,
    for example: get the values of the property P31 for the entity Q104534511;
    query_results: dict, Wikidata response in json format;
    QID: str, only 1 QID;
    properties: list of str; for example ['P31','P279'];
    Return a dict {'QID': {'property_ID':{'property_value': ''}}};
    The empty value of the property_value is intended for its label, which is queried using another function
    """
    dict_per_qid = {}
    dict_per_qid[QID] = {}
    
    for p in propepties:

        dict_per_qid[QID][p] = {}
        if 'missing' not in query_results['entities'][QID].keys():
            claims_per_qid = query_results['entities'][QID]['claims']
            # getting property values
            if p in claims_per_qid:
                for i in claims_per_qid[p]:
                    if 'datavalue' in i['mainsnak']:
                        dict_per_qid[QID][p][i['mainsnak']['datavalue']['value'].get('id')] = ""
                    
    return dict_per_qid

def get_claims(qids:list, propepties:list, user_agent:str, path='') -> str:

    """
    Getting claims by QIDs and properties
    qids: list of entities (str) to work on, if len > 50, results will be saved in batches (separate files)
    properties: list of str; for example ['P31','P279']; getting claims only with the properties specified in this list
    user_agent: str with user-agent's info; required by Wikidata (see: https://meta.wikimedia.org/wiki/User-Agent_policy)
    path: optional; str path where to save json files (including '/');
    Saves the request results in json files due to large size;
    Returns str when files are saved and whether there are failed entities;
    Saves failed entities (due to request errors) in a separate txt file;
    Uses another function (_claims_parse_properties) to parse the indicated properties from the request results
    """

    # 'wbgetentities' constant params
    url = "https://www.wikidata.org/w/api.php"
    params = {"action":"wbgetentities",
              "ids":"", # string of entities (max=50)
              "props":"claims",
              "format":"json"}
    headers = {"user-agent":user_agent}

    failed_entities = []

    # - N LOOPS - #

    # if there's a remainder
    if len(qids) % 50 > 0:
        loops = len(qids) // 50 + 1 # add another loop for requests
    else:
        loops = len(qids) // 50

    # - REQUEST LOOPS - #   

    # counters to slice qids

    start_quid_str = 0
    end_quid_str = 0

    for i in range(0,loops):
        
        results = {}

        ids_string = "" # putting Qs in one string
        end_quid_str = end_quid_str + 50 # max 50 entities per request

        for q in qids[start_quid_str:end_quid_str]:
            ids_string = ids_string + f"{q}|"

        start_quid_str = start_quid_str + 50

        # updating params
        params["ids"] = ids_string.rstrip("|")

        # sending a request
        d = requests.get(url,params=params,headers=headers)
        claims = d.json() # claims per request

        if 'entities' in claims:
            for entity in claims['entities']:
                results.update(_claims_parse_properties(claims,entity,propepties))

            # - SAVING RESULTS IN BATCHES - #
    
            with open(f'{path}claims_{start_quid_str}.json', 'w') as json_file:
                json.dump(results, json_file)
            # to prevent server errors
            time.sleep(2)

        else:
            failed_entities.extend(qids[start_quid_str:end_quid_str])

    if len(failed_entities) > 0:
        with open(f'{path}failed_entities.txt','w') as txt_file:
            for q in failed_entities:
                txt_file.writelines(f"{q}\n")
        output_str = f"Requests are saved in {path}. There are {len(failed_entities)} failed entities, see the .txt file"
    else:
        output_str = f"Requests are saved in {path}"

    return output_str

def find_where_query_term_appears(query_results_path:str, lang:str) -> dict:
    """
    Finding where a target term appears in the search results: label, description, or aliases
    query_results_path: str, the path to the search results  (the output of 'get_lit_by_term'), including '/'
    lang – str language string ("en" or "nl") corresponding to the results language of 'get_lit_by_term'
    Returns dict with info about every hit and where a term was found
    """
    
    # importing query results
    with open(query_results_path,'r') as jf:
        query_results = json.load(jf)

    target_terms_dict_wikidata = {}

    for query_term, hits in query_results.items():
        
        results_per_query_term = []

        for page_id, values in hits.items():
            if 'entityterms' in values:
                
                # checking labels
                if 'label' in values['entityterms']:
                    if len(re.findall(f'\\b{query_term}\\b',values['entityterms']['label'][0],re.IGNORECASE)) > 0:
                        dict_per_hit = {}
                        dict_per_hit['query_term'] = query_term
                        dict_per_hit['lang'] = lang
                        dict_per_hit['QID'] = values['title']
                        dict_per_hit['prefLabel'] = values['entityterms']['label'][0]
                        dict_per_hit['aliases'] = values['entityterms'].get('alias')
                        dict_per_hit['description'] = values['entityterms'].get('description')
                        dict_per_hit['index'] = values['index']
                        dict_per_hit['found_in'] = 'prefLabel'
                        results_per_query_term.append(dict_per_hit)
                        
                # checking aliases        
                if 'alias' in values['entityterms']:
                    for a in values['entityterms']['alias']:
                        if len(re.findall(f'\\b{query_term}\\b',a,re.IGNORECASE)) > 0:
                            dict_per_hit = {}
                            dict_per_hit['query_term'] = query_term
                            dict_per_hit['lang'] = lang
                            dict_per_hit['QID'] = values['title']
                            if values['entityterms'].get('label') != None:
                                dict_per_hit['prefLabel'] = values['entityterms']['label'][0]
                            else:
                                dict_per_hit['prefLabel'] = None
                            dict_per_hit['aliases'] = values['entityterms']['alias']
                            dict_per_hit['description'] = values['entityterms'].get('description')
                            dict_per_hit['index'] = values['index']
                            dict_per_hit['found_in'] = 'aliases'
                            results_per_query_term.append(dict_per_hit)
                            
                # checking descriptions            
                if 'description' in values['entityterms']:
                    if len(re.findall(f'\\b{query_term}\\b',values['entityterms']['description'][0],re.IGNORECASE)) > 0:
                        dict_per_hit = {}
                        dict_per_hit['query_term'] = query_term
                        dict_per_hit['lang'] = lang
                        dict_per_hit['QID'] = values['title']
                        if values['entityterms'].get('label') != None:
                            dict_per_hit['prefLabel'] = values['entityterms']['label'][0]
                        else:
                            dict_per_hit['prefLabel'] = None
                        dict_per_hit['aliases'] = values['entityterms'].get('alias')
                        dict_per_hit['description'] = values['entityterms']['description'][0]
                        dict_per_hit['index'] = values['index']
                        dict_per_hit['found_in'] = 'description'
                        results_per_query_term.append(dict_per_hit)
                        
        target_terms_dict_wikidata[query_term] = results_per_query_term

    return target_terms_dict_wikidata


def filter_proper_names(dataset_path:str, lang:str, path:'') -> str:
    """
    Filtering out proper names in the search results
    dataset_path: str, the path to the search results with claims ('results_w_claims_[lang].json')
    lang: str, the lang of the search results ('en' or 'nl')
    path: optional; str path where to save json files (including '/');
    Saves a json file with search results without proper names (results_clean_[lang].json)
    Returns a status str with N of filtered out entities
    """

    # importing results w claims
    with open(dataset_path,'r') as jf:
        dataset = json.load(jf)

    results_clean = {}
    n_filtered = 0

    if lang == 'en':
        Q5_value = 'human'
    if lang == 'nl':
        Q5_value = 'mens'

    for query_term,results in dataset.items():
        results_clean[query_term] = []
        for hit in results:
            count_Q5 = 0
            for p in hit['instance_of']:
                    if p == Q5_value:
                        count_Q5 += 1

            if hit['prefLabel'] != None and len(re.findall(f"\\b{hit['query_term'].capitalize()}\\b",hit['prefLabel'])) > 0 and count_Q5 > 0:
                n_filtered += 1

            else:
                results_clean[query_term].append(hit)

    with open(f"{path}results_clean_{lang}.json","w") as jf:
        json.dump(results_clean, jf)
                
    return f"File is saved. {n_filtered} entities are filtered out."


def get_bows(path_to_results:str, lang:str) -> dict:
    '''
    Getting bag of words (BoW) from the Wikidata search results for every search term
    path_to_results: str, a path to the search results (in json format)
    lang: str, 'en' or 'nl'
    Returns a dict with BoWs per hit per term: {term:[{QID:['token1','token2','token3']}]}
    '''
    wnl = WordNetLemmatizer()
    all_bows = {}
   
    with open(path_to_results,'r') as jf:
        search_results = json.load(jf)
    
    for query_term, results in search_results.items():
        list_by_term = []

        for hit in results:
            q_bag = {}
            literals = []
            
            if hit["prefLabel"] != None:
                literals.append(hit["prefLabel"]) # adding prefLabel

            if hit["found_in"] == "prefLabel" or hit["found_in"] == "aliases":
                literals.extend(hit["instance_of"])
                literals.extend(hit["subclass_of"])

            if hit["aliases"] != None:
                literals.extend(hit["aliases"]) # adding aliases

            # adding descriptions
            # descriptions can be str or list
            # None doesn't add
            if type(hit["description"]) == str:
                literals.append(hit["description"])
                
            if type(hit["description"]) == list:
                literals.extend(hit["description"])
            #print(literals)    
            bow = bows.make_bows(literals,lang,merge_bows=True)
                
            q_bag[hit["QID"]] = bow

            list_by_term.append(q_bag)

        all_bows[query_term] = list_by_term
        
    return all_bows


def get_lit_related_matches_bow(lang:str) -> dict:
    '''
    Generates dict with BoWs by query terms and their related matches in Wikidata
    reads files with Wikidata BoWs: gzip_wd_bows_en.json (EN), gzip_wd_bows_nl.json (NL)
    the info about which QID is a related match to the query term is taken from https://github.com/cultural-ai/wordsmatter/blob/main/related_matches/rm.csv
    lang: str, 'en' or 'nl'
    Returns a dict with BoWs by term: {term:{QID:['token1','token2','token3']}}
    '''
    
    results = {}
    
    # loading related matches
    path_rm = "https://github.com/cultural-ai/wordsmatter/raw/main/related_matches/rm.json"
    rm = requests.get(path_rm).json()
    
    # load wd bows
    path_to_bows = f"https://github.com/cultural-ai/LODlit/raw/main/Wikidata/gzip_wd_bows_{lang}.json"
    wd_bow = requests.get(path_to_bows).json()

    # getting a list of all QIDs of related matches in Wikidata
    related_matches_wd_qid = list(set([values["related_matches"]["wikidata"][0] for values in rm.values() \
                     if values["lang"] == lang and values["related_matches"]["wikidata"][0] != 'None']))    
    # getting BoWs for QIDs
    related_matches_wd_qid_bows = {}
    for q_id_rm in related_matches_wd_qid:
        for hits in wd_bow.values():
            for hit in hits:
                for q_id, bow in hit.items():
                    if q_id == q_id_rm:
                        related_matches_wd_qid_bows[q_id] = bow
                        
    # shaping resulting dict: terms with related matches and BoWs
    for values in rm.values():
        if values["lang"] == lang:
            rm_wd = values["related_matches"]["wikidata"][0]
            if rm_wd != "None":
                for term in values["query_terms"]:
                    if rm_wd in related_matches_wd_qid_bows.keys():
                        results[term] = {"QID":rm_wd,"bow":related_matches_wd_qid_bows[rm_wd]}
            
    return results

def get_cs(lang:str):
    '''
    Calculating three cosine similarity scores between Wikidata search results and background info;
    the similarity scores are based on (1) only related matches, (2) only WM text, and (3) extended bows with related matches and WM text
    lang:str, language of bows, "en" or "nl"
    Returns a pandas data frame with columns:
    query_term, QID, bow, cs_rm, cs_wm, cs_rm_wm
    '''

    nlp = bows.load_spacy_nlp(lang)

    # load bckground info
    path_bg = "https://github.com/cultural-ai/LODlit/raw/main/bg/background_info_bows.json"
    bg_info = requests.get(path_bg).json()

    wikidata_df = pd.DataFrame(columns=['term','hit_id','bow','cs_rm','cs_wm','cs_rm_wm'])

    # load wd bows
    path_wd_bows = f"https://github.com/cultural-ai/LODlit/raw/main/Wikidata/gzip_wd_bows_{lang}.json"
    wikidata_bows = requests.get(path_wd_bows).json()

    for term, hits in wikidata_bows.items():

        bg_rm = bows._collect_bg(term,lang,bg_info,bg_bow="rm")
        bg_wm = bows._collect_bg(term,lang,bg_info,bg_bow="wm")
        bg_rm_wm = bows._collect_bg(term,lang,bg_info,bg_bow="joint")
          
        # if there are search results
        if len(hits) > 0:
            for hit in hits:
                for i, bow in hit.items():
                    # making a set
                    wd_bow = list(set(bow))
                    
                    if len(wd_bow) > 0:
                        # calculate cs
                        cs_rm = bows.calculate_cs(bg_rm,wd_bow,nlp)
                        cs_wm = bows.calculate_cs(bg_wm,wd_bow,nlp)
                        cs_rm_wm = bows.calculate_cs(bg_rm_wm,wd_bow,nlp)
                        
                        wikidata_df.loc[len(wikidata_df)] = [term,i,wd_bow,cs_rm,cs_wm,cs_rm_wm]
                    
                    # if there are no tokens, all cs = None
                    else:
                        wikidata_df.loc[len(wikidata_df)] = [term,i,wd_bow,None,None,None]

        # if there are no search results in Wikidata, cs = None
        else:
            wikidata_df.loc[len(wikidata_df)] = [term,None,None,None,None,None]

    return wikidata_df


def get_n_hits_by_properties(path_to_results:str, lang:str, group_by_lemma=False):
    '''
    Getting N of hits of query terms by properties in Wikidata
    with optional grouping by lemmas
    path_to_results: str, path to a json file with results (query results or annotated entities)
    lang: str, language of results, "en" or "nl"
    group_by_lemma: bool, if group N hits by lemma, default False (count by query term)
    Requires the file with query terms "query_terms.json"
    Returns a pandas dataframe
    '''
    # importing query terms with lemmas
    # loading query terms
    path_query_terms = "https://github.com/cultural-ai/LODlit/raw/main/query_terms.json"
    query_terms = requests.get(path_query_terms).json()
    
    with open(path_to_results,'r') as jf:
        wd_results = json.load(jf)
        
    n_occurences_by_query_term = pd.DataFrame(columns=["lemma","query_term","lang","wd_prefLabel","wd_aliases","wd_descr","wd_qt_total"])

    for term, hits in wd_results.items():
        # property counters
        pref = 0
        alias = 0
        descr = 0
        
        for hit in hits:
            if hit["found_in"] == "prefLabel":
                pref += 1
            if hit["found_in"] == "aliases":
                alias += 1
            if hit["found_in"] == "description":
                descr += 1
        
        total_count_by_qt = pref + alias + descr
                
        # getting lemma of a query term
        for l, wordforms in query_terms[lang].items():
                if term in wordforms:
                    lemma = l
        
        data = [lemma,term,lang,pref,alias,descr,total_count_by_qt]
        n_occurences_by_query_term.loc[len(n_occurences_by_query_term)] = data
        
        results_to_return = n_occurences_by_query_term
                
    if group_by_lemma == True:
        
        n_occurences_by_lemma = pd.DataFrame(columns=["lemma","lang","wd_prefLabel","wd_aliases","wd_descr","wd_lemma_total"])
        
        for group in n_occurences_by_query_term.groupby('lemma'):
            row = [group[0],lang,sum(group[1]['wd_prefLabel']),sum(group[1]['wd_aliases']),\
              sum(group[1]['wd_descr']),sum(group[1]['wd_qt_total'])]
            n_occurences_by_lemma.loc[len(n_occurences_by_lemma)] = row
            
        results_to_return = n_occurences_by_lemma
        
    return results_to_return



# Open Dutch WordNet parser
# Download Open Dutch WordNet from "https://github.com/cultural-ai/OpenDutchWordnet"
# requires lxml: pip install lxml

import sys
import json
import csv
import re
import requests
import pandas as pd
import nltk
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer
import simplemma
import bows

def main():
    if __name__ == "__main__":
        main()

def _set_odwn(path_odwn:str):
    """
    path_odwn: str path to the directory with OpenDutchWordnet (not including the module itself, for example "user/Downloads")
    Inserts the path to the OpenDutchWordnet module
    Download OpenDutchWordnet from "https://github.com/cultural-ai/OpenDutchWordnet"
    Returns an instance using Wn_grid_parser
    """

    sys.path.insert(0,path_odwn)
    from OpenDutchWordnet import Wn_grid_parser
    # creating an instance
    odwn_instance = Wn_grid_parser(Wn_grid_parser.odwn)

    return odwn_instance

def get_le_info(le_ids:list, path_odwn:str) -> dict:
    """
    Getting Lemma, sense definitions, examples, synset ID, synset definitions of ODWN Lexical Entries
    le_ids: list of Lexical Entries IDs (str), for example "appel-n-1"
    path_odwn: str path to the directory with OpenDutchWordnet (not including the module itself, for example "user/Downloads")
    Returns a dict: {"le_id": {"lemma": "",
                                "sense_def": "",
                                "examples": [],
                                "synset_ID": "",
                                "synset_def": []}
    """
    # importing ODWN
    instance = _set_odwn(path_odwn)
    
    # load all glosses
    path_to_glosses = "https://github.com/cultural-ai/LODlit/raw/main/ODWN/odwn_synset_glosses.json"
    synset_glosses = requests.get(path_to_glosses).json()
    
    results_odwn = {}

    for le_id in le_ids:
        le = instance.les_find_le(le_id)
        synset_id = le.get_synset_id()

        sense_def = le.get_definition()
        if sense_def == "":
            sense_def = None

        results_odwn[le_id] = {"lemma": le.get_lemma(),
                                "sense_def": sense_def,
                                "examples": le.get_sense_example(),
                                "synset_ID": synset_id,
                                "synset_def": synset_glosses.get(synset_id)}
    return results_odwn

def _shape_search_results(instance,query_term:str,le,all_synset_definitions:dict,found_in:str,example="") -> dict:
    
    """
    instance: class wn_grid_parser.Wn_grid_parser
    query_term: str
    le: class le.Le
    all_synset_definitions: dict, parses "https://github.com/cultural-ai/LODlit/raw/main/ODWN/odwn_synset_glosses.json" 
    found_in: str
    example: str, default is ""
    Returns a dict of search results
    This function does not perform searching,
    but puts the search results of "find_terms" in a dict
    """
    
    result_dict = {}
    result_dict["query_term"] = query_term
    result_dict["le_id"] = le.get_id()
    result_dict["le_written_form"] = le.get_lemma()
    result_dict["sense_id"] = le.get_sense_id()
    result_dict["sense_definition"] = le.get_definition()
    result_dict["sense_examples"] = le.get_sense_example()

    if le.get_synset_id() != None:
        synset_id = le.get_synset_id()
        result_dict["synset_id"] = synset_id
        result_dict["synonyms"] = [les.get_lemma() for les in instance.les_all_les_of_one_synset(synset_id)]
        if synset_id in all_synset_definitions.keys():
            result_dict["synset_definitions"] = all_synset_definitions[synset_id]
        else:
            result_dict["synset_definitions"] = []
    else:
        result_dict["synset_id"] = ""
        result_dict["synonyms"] = []
        result_dict["synset_definitions"] = []
        
    result_dict["found_in"] = found_in
    
    if found_in == "sense_examples":
        # there can be multiple examples; writing the example the term was found in
        result_dict["found_in_example"] = example
    
    return result_dict

def find_terms(query_terms:list, path_odwn:str) -> dict:
    '''
    Searching for a query term in lemmas (written form of Lexical Entry (LE)), sense definitions, sense examples, and synset definitions (glosses) of ODWN
    query_term: list of terms to search in ODWN (str), phrases are also allowed
    path_odwn: str path to the directory with OpenDutchWordnet (not including the module itself, for example "user/Downloads")
    Returns a dict with the LE or synset info by query term if the query term was found
    if found in LE or its children tags:
        {query_term:[{"query_term": "",
                    "le_id": "",
                   "le_written_form": "",
                   "sense_id": "",
                   "sense_definition": "",
                   "sense_examples": [],
                   "synset_id": "",
                   "synonyms": [],
                   "synset_definitions": [],
                   "found_in": ""}]};
    if found in a synset definition, but that synset is not attached to any LEs:
        {query_term:[{"query_term": "",
                   "synset_id": "",
                   "synonyms": [],
                   "synset_definitions": [],
                   "found_in": "synset_definitions",
                   "found_in_synset_definition": ""}]}
    '''

    # importing ODWN
    instance = _set_odwn(path_odwn)
    
    # load all glosses
    path_to_glosses = "https://github.com/cultural-ai/LODlit/raw/main/ODWN/odwn_synset_glosses.json"
    synset_glosses = requests.get(path_to_glosses).json()

    results = {}

    for query_term in query_terms:

        results_per_term = []

        # searching in lemmas
        for le in instance.lemma_get_generator(query_term,ignore_case=True):
            results_per_term.append(_shape_search_results(instance,query_term,le,synset_glosses,"le"))

        # Iterating over all Lexical Entries
        
        # searching in sense definitions
        for le in instance.les_get_generator():
            if len(re.findall(f"\\b{query_term}\\b",le.get_definition(),re.IGNORECASE)) > 0:
                results_per_term.append(_shape_search_results(instance,query_term,le,synset_glosses,"sense_definition"))

            # searching in sense examples
            for example in le.get_sense_example():
                if len(re.findall(f"\\b{query_term}\\b",example,re.IGNORECASE)) > 0:
                    results_per_term.append(_shape_search_results(instance,query_term,le,synset_glosses,"sense_examples",example))

        # searching in synset definitions
        for synset_id, definitions in synset_glosses.items():
            for d in definitions:
                if len(re.findall(f"\\b{query_term}\\b",d,re.IGNORECASE)) > 0:
                        # results_per_term for synsets are different, so we don"t use "_shape_search_results()"
                        result_dict = {}
                        result_dict["query_term"] = query_term
                        result_dict["synset_id"] = synset_id
                        result_dict["synonyms"] = [les.get_lemma() for les in instance.les_all_les_of_one_synset(synset_id)]
                        result_dict["synset_definitions"] = synset_glosses[synset_id]
                        result_dict["found_in"] = "synset_definitions"
                        # there can be multiple definitions; writing the definition the term was found in
                        result_dict["found_in_synset_definition"] = d
                        results_per_term.append(result_dict)

        results[query_term] = results_per_term

    return results

def get_bows(path_to_results:str) -> dict:
    '''
    Getting bag of words (BoW) from the Open Dutch Wordnet search results for every search term
    path_to_results: str, a path to the search results (in json format)
    Returns a dict with BoWs per hit per term: {term:[{synset_id/le_id:['token1','token2','token3']}]}
    '''
    wnl = WordNetLemmatizer()

    all_bows = {}
   
    with open(path_to_results,'r') as jf:
        search_results = json.load(jf)

    for query_term, results in search_results.items():

        list_by_term = []

        for hit in results:

            q_bag = {}
            literals = []

            if hit["found_in"] == "synset_definitions":
                literals.extend(hit["synonyms"])
                literals.extend(hit["synset_definitions"])
            else:
                literals.append(hit["le_written_form"])
                literals.extend(hit["sense_examples"])
                literals.extend(hit["synonyms"])
                literals.extend(hit["synset_definitions"])
                if hit["sense_definition"] != "":
                    literals.append(hit["sense_definition"])

            bow = bows.make_bows(literals,"nl",merge_bows=True)

            # if there's no synset ID, using LE ID as key
            if hit["synset_id"] != "":
                q_bag[hit["synset_id"]] = bow
            else:
                q_bag[hit["le_id"]] = bow

            list_by_term.append(q_bag)

        all_bows[query_term] = list_by_term

    return all_bows


def get_lit_related_matches_bow() -> dict:
    '''
    Generates dict with BoWs by query terms and their related matches in ODWN
    reads two json files: (1) odwn_bows.json (ODWN BoWs), (2) related_matches_odwn.json (related matches for every term in ODWN)
    Returns a dict with BoWs by term: {term:['token1','token2','token3']}
    This dict doesn't store info about synsets IDs or LE IDs bc one term can have one or more related matches;
    in these cases, BoWs of each synsets and LEs are merged
    '''
    
    results = {}
    
    # loading related matches
    path_rm = "https://github.com/cultural-ai/wordsmatter/raw/main/related_matches/rm.json"
    rm = requests.get(path_rm).json()

    # load odwn bows
    path_to_bows = f"https://github.com/cultural-ai/LODlit/raw/main/ODWN/odwn_bows.json"
    odwn_bows = requests.get(path_to_bows).json()
    
    # getting a list of all ODWN LE IDs and synset IDs of related matches
    odwn_les_synsets = []
    for values in rm.values():
        if values["odwn_le"][0] != 'None':
            odwn_les_synsets.extend(values["odwn_le"])
        if values["odwn_synsets"] != '':
            odwn_les_synsets.extend(values["odwn_synsets"])

    related_matches_odwn = list(set(odwn_les_synsets))
            
    # getting BoWs for each LE or synset
    related_matches_odwn_bows = {}
    for odwn_id in related_matches_odwn:
        for hits in odwn_bows.values():
            for hit in hits:
                for le, bow in hit.items():
                    if le == odwn_id:
                        related_matches_odwn_bows[le] = bow
                        
    # shaping resulting dict: terms with related matches and BoWs
    for values in rm.values():
        rm_ids = []
        rm_ids.extend(values["odwn_le"])
        rm_ids.extend(values["odwn_synsets"])
        if rm_ids != []:
            for term in values["query_terms"]:
                bow = []
                for i in rm_ids:
                    if i in related_matches_odwn_bows.keys():
                        bow.extend(related_matches_odwn_bows[i])
                        results[term] = bow
    
    return results

def get_cs():
    '''
    Calculating three cosine similarity scores between ODWN search results and background info;
    the similarity scores are based on (1) only related matches, (2) only WM text, and (3) extended bows with related matches and WM text
    Returns a pandas data frame with columns:
    query_term, hit_id, bow, cs_rm, cs_wm, cs_rm_wm
    '''

    nlp = bows._load_spacy_nlp("nl")

    # load bckground info
    path_bg = "https://github.com/cultural-ai/LODlit/raw/main/bg/background_info_bows.json"
    bg_info = requests.get(path_bg).json()

    # load all pwn bows
    path_odwn_bows = f"https://github.com/cultural-ai/LODlit/raw/main/ODWN/odwn_bows.json"
    odwn_bows = requests.get(path_pwn_bows).json()

    odwn_df = pd.DataFrame(columns=['term','hit_id','bow','cs_rm','cs_wm','cs_rm_wm'])

    for term, hits in odwn_bows.items():

        bg_rm = bows._collect_bg(term,"nl",bg_info,bg_bow="rm")
        bg_wm = bows._collect_bg(term,"nl",bg_info,bg_bow="wm")
        bg_rm_wm = bows._collect_bg(term,"nl",bg_info,bg_bow="joint")

        # if there are search results
        if len(hits) > 0:
            for hit in hits:
                for i, bow in hit.items():
                    # making a set
                    bow_set = list(set(bow))

                    if len(bow_set) > 0:
                        # calculate cs
                        cs_rm = bows.calculate_cs(bg_rm,bow_set,nlp)
                        cs_wm = bows.calculate_cs(bg_wm,bow_set,nlp)
                        cs_rm_wm = bows.calculate_cs(bg_rm_wm,bow_set,nlp)
                        
                        odwn_df.loc[len(odwn_df)] = [term,i,bow_set,cs_rm,cs_wm,cs_rm_wm]
                    
                    # if there are no tokens, all cs = None
                    else:
                        odwn_df.loc[len(odwn_df)] = [term,i,bow_set,None,None,None]

        # if there are no search results in PWN, cs = None
        else:
            odwn_df.loc[len(odwn_df)] = [term,None,None,None,None,None]

    return odwn_df

def get_pragmatics(ids:list, path_odwn:str, by="synset_id") -> dict:
    '''
    Get pragmatic values of lemmas by synset IDs or sense IDs
    ids: list, synset ids or sense_ids, for example ["temperatuur-n-4", "temperatuur-n-3"]
    path_odwn: str, path to the directory with OpenDutchWordnet (not including the module itself, for example "user/Downloads")
    by: str, how to get pragmatics: by "synset_id" or "sense_id", default is "synset_id"
    Returns dict: {'id': [{'le_id': 'lemma_id',
                       'pragmatics': {'register': None,
                        'geography': None,
                        'chronology': None,
                        'connotation': None,
                        'domain': None}}]}
    '''
    # importing ODWN
    instance = _set_odwn(path_odwn)

    pragmatics = {}

    for i in ids:

        pragmatics_per_le = []

        if by == "synset_id":

            for le in instance.les_all_les_of_one_synset(i):
                dict_per_le = {}
                dict_per_le["le_id"] = le.get_id()
                dict_per_le["pragmatics"] = le.get_pragmatics()
                pragmatics_per_le.append(dict_per_le)

        if by == "sense_id":

            for le in instance.les_get_generator():
                if le.get_sense_id() == i:
                    dict_per_le = {}
                    dict_per_le["le_id"] = le.get_id()
                    dict_per_le["pragmatics"] = le.get_pragmatics()
                    pragmatics_per_le.append(dict_per_le)

        pragmatics[i] = pragmatics_per_le

    return pragmatics